from __future__ import annotations

import tempfile
from unittest.mock import patch

import pytest
from google.api_core import page_iterator
from google.cloud import storage
from google.cloud.exceptions import Conflict, NotFound

from mockgcp import mock_storage
from mockgcp.storage.bucket import MockBucket
from mockgcp.storage.client import MockClient


class TestBucketConstructor:
    @mock_storage
    def test_with_valid_name(self) -> None:
        client = storage.Client()
        assert isinstance(client, MockClient)

        bucket = client.bucket("test-bucket-name")

        assert isinstance(bucket, MockBucket)
        assert bucket.name == "test-bucket-name"

    @mock_storage
    def test_with_invalid_name(self) -> None:
        client = storage.Client()
        with pytest.raises(
            ValueError, match="Bucket names must start and end with a number or letter."
        ):
            client.bucket("test-bucket-name-")


class TestGetBucket:
    @mock_storage
    def test_with_existing_bucket_name(self) -> None:
        client = storage.Client()
        bucket = client.create_bucket("test-bucket-name")

        assert client.get_bucket("test-bucket-name") is bucket

    @mock_storage
    def test_with_non_existing_bucket_name(self) -> None:
        client = storage.Client()
        with pytest.raises(NotFound):
            client.get_bucket("test-bucket-name")


class TestLookupBucket:
    @mock_storage
    def test_wiht_existing_bucket_name(self) -> None:
        client = storage.Client()
        bucket = client.create_bucket("test-bucket-name")

        assert client.lookup_bucket("test-bucket-name") is bucket

    @mock_storage
    def test_wiht_non_existing_bucket_name(self) -> None:
        client = storage.Client()

        assert client.lookup_bucket("test-bucket-name") is None


class TestCreateBucket:
    @mock_storage
    def test_simple(self) -> None:
        client = storage.Client()
        bucket = client.create_bucket("test-bucket-name")

        assert isinstance(bucket, MockBucket)
        assert list(client.list_buckets()) == [bucket]

    @mock_storage
    def test_with_existing_bucket_name(self) -> None:
        client = storage.Client()
        bucket = client.create_bucket("test-bucket-name")

        with pytest.raises(Conflict):
            client.create_bucket("test-bucket-name")

        with pytest.raises(Conflict):
            bucket.create()


class TestListBlobs:
    @mock_storage
    def test_simple(self) -> None:
        client = storage.Client()
        bucket = client.create_bucket("test-bucket-name")
        bucket.list_blobs()


class TestListBuckets:
    @mock_storage
    def test_with_no_bucket(self) -> None:
        client = storage.Client()
        buckets = client.list_buckets()

        assert isinstance(buckets, page_iterator.HTTPIterator)
        assert list(buckets) == []

    @mock_storage
    def test_with_one_bucket(self) -> None:
        client = storage.Client()
        bucket = client.create_bucket("test-bucket-name")
        buckets = client.list_buckets()

        assert list(buckets) == [bucket]

    @mock_storage
    def test_with_max_results(self) -> None:
        client = storage.Client()
        client.create_bucket("test-bucket-name-n1")
        client.create_bucket("test-bucket-name-n2")

        assert len(list(client.list_buckets(max_results=1))) == 1

    @mock_storage
    def test_with_prefix(self) -> None:
        client = storage.Client()
        bucket_test = client.create_bucket("test-bucket-name")
        client.create_bucket("other-bucket-name")

        assert list(client.list_buckets(prefix="test")) == [bucket_test]


class TestDeleteBucket:
    @mock_storage
    def test_delete_bucket(self) -> None:
        client = storage.Client()
        bucket_test = client.create_bucket("test-bucket-name")
        bucket_test.delete()


class TestBlobs:
    @classmethod
    def setup_class(cls):
        cls.patcher = patch("google.cloud.storage.Client", MockClient)
        cls.patcher.start()
        cls.client = storage.Client()
        cls.bucket = cls.client.create_bucket("test-bucket-name")
        cls.file = tempfile.NamedTemporaryFile()
        cls.file.write(b"test")

        cls.new_file = tempfile.NamedTemporaryFile()

    @classmethod
    def teardown_class(cls) -> None:
        cls.bucket.delete()
        cls.file.close()
        cls.new_file.close()

        del cls.client
        cls.patcher.stop()

    class TestFilename:
        def test_upload(self) -> None:
            blob = self.bucket.blob("dir/dir/dir/file.bin")
            blob.upload_from_filename(self.file.name)

        def test_download(self) -> None:
            blob = self.bucket.blob("dir/dir/dir/file.bin")
            blob.download_to_filename(self.new_file.name)
            assert self.new_file.read() == self.file.read()

    class TestText:
        def test_upload(self) -> None:
            blob = self.bucket.blob("dir/dir/dir/file.txt")
            blob.upload_from_string("Some text")

        def test_download(self) -> None:
            blob = self.bucket.blob("dir/dir/dir/file.txt")
            assert blob.download_as_string() == "Some text"



from __future__ import annotations

from mockgcp.storage.client import MockClient
from mockgcp.storage.blob import MockBlob
from mockgcp.storage.bucket import MockBucket

__version__ = "0.2.0"

# List of public objects in storage module
__all__ = ["MockBlob", "MockBucket", "MockClient", "__version__"]



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, NoReturn
from unittest import mock
from urllib.parse import urlsplit

from google.api_core import page_iterator
from google.cloud.exceptions import NotFound
from google.cloud.storage._helpers import _PropertyMixin, _validate_name
from google.cloud.storage.constants import _DEFAULT_TIMEOUT
from google.cloud.storage.retry import (
    DEFAULT_RETRY,
    DEFAULT_RETRY_IF_GENERATION_SPECIFIED,
)

from mockgcp.storage.blob import MockBlob

if TYPE_CHECKING:
    from mockgcp.storage.client import MockClient


class MockBucket(_PropertyMixin):
    def __init__(self, client: MockClient, name=None, user_project=None) -> None:
        name = _validate_name(name)
        super().__init__(name)
        self._client = client
        self._user_project = user_project

    def __repr__(self) -> str:
        return f"<Bucket: {self.name}>"

    @property
    def client(self) -> MockClient:
        return self._client

    @property
    def blobs(self):
        return self.client.backend.blobs[self.name]

    @property
    def data_dir(self) -> Path:
        return self.client.backend.data_dir / self.name

    @property
    def user_project(self):
        return self._user_project

    @classmethod
    def from_string(cls, uri, client=None):
        scheme, netloc, path, query, frag = urlsplit(uri)

        if scheme != "gs":
            msg = "URI scheme must be gs"
            raise ValueError(msg)

        return cls(client, name=netloc)

    def blob(
        self,
        blob_name,
        chunk_size=None,
        encryption_key=None,
        kms_key_name=None,
        generation=None,
    ):
        return MockBlob(
            blob_name, self, chunk_size, encryption_key, kms_key_name, generation
        )

    def notification(
        self,
        topic_name,
        topic_project=None,
        custom_attributes=None,
        event_types=None,
        blob_name_prefix=None,
        # payload_format=NONE_PAYLOAD_FORMAT,
    ) -> NoReturn:
        raise NotImplementedError

    def exists(self, client=None):
        return self.name in self.client.buckets

    def create(
        self,
        client=None,
        project=None,
        location=None,
        predefined_acl=None,
        predefined_default_object_acl=None,
        enable_object_retention=False,
        timeout=_DEFAULT_TIMEOUT,
        retry=DEFAULT_RETRY,
    ) -> None:
        client = self._require_client(client)
        client.create_bucket(
            self.name,
            project=project,
            location=location,
            predefined_acl=predefined_acl,
            predefined_default_object_acl=predefined_default_object_acl,
            enable_object_retention=enable_object_retention,
            timeout=timeout,
            retry=retry,
        )

    def patch(self, client=None) -> NoReturn:
        raise NotImplementedError

    @property
    def acl(self) -> NoReturn:
        raise NotImplementedError

    @property
    def default_object_acl(self) -> NoReturn:
        raise NotImplementedError

    @staticmethod
    def path_helper(bucket_name):
        return "/b/" + bucket_name

    @property
    def path(self):
        if not self.name:
            msg = "Cannot determine path without bucket name."
            raise ValueError(msg)

        return self.path_helper(self.name)

    def get_blob(
        self, blob_name, client=None, encryption_key=None, generation=None, **kwargs
    ):
        return self.client.backend.blobs.get(self.name, {}).get(blob_name)

    def list_blobs(
        self,
        max_results=None,
        page_token=None,
        prefix=None,
        delimiter=None,
        start_offset=None,
        end_offset=None,
        include_trailing_delimiter=None,
        versions=None,
        projection="noAcl",
        fields=None,
        client=None,
        timeout=_DEFAULT_TIMEOUT,
        retry=DEFAULT_RETRY,
        match_glob=None,
        include_folders_as_prefixes=None,
        soft_deleted=None,
        page_size=None,
    ):
        if isinstance(max_results, int):
            blobs = list(self.client.backend.blobs[self.name].values())[:max_results]
        else:
            blobs = list(self.client.backend.blobs[self.name].values())

        if isinstance(delimiter, str):
            raise NotImplementedError

        if isinstance(prefix, str):
            blobs = [
                blob for blob in blobs if Path(blob.name).is_relative_to(Path(prefix))
            ]

        extra_params = {"projection": projection}

        path = "/foo"
        page_response = {"items": blobs}
        api_request = mock.Mock(return_value=page_response)

        iterator = page_iterator.HTTPIterator(
            mock.sentinel.client,
            api_request,
            path=path,
            item_to_value=page_iterator._item_to_value_identity,
            max_results=max_results,
            page_token=mock.sentinel.token,
            extra_params=extra_params,
        )
        iterator.prefixes = set()
        iterator.bucket = self
        return iterator

    def delete(self):
        for blob in list(self.blobs.values()):
            blob.delete()
        del self.client.buckets[self.name]

    def delete_blob(
        self,
        blob_name,
        client=None,
        generation=None,
        if_generation_match=None,
        if_generation_not_match=None,
        if_metageneration_match=None,
        if_metageneration_not_match=None,
        timeout=_DEFAULT_TIMEOUT,
        retry=DEFAULT_RETRY_IF_GENERATION_SPECIFIED,
    ) -> None:
        if blob_name in self.blobs:
            blob = self.blobs.pop(blob_name)
            blob.file_name.unlink(missing_ok=True)
        else:
            msg = f"404 GET https://storage.googleapis.com/storage/v1/b/{self.name}/o/{blob_name}?projection=noAcl"
            raise NotFound(msg)



from __future__ import annotations

import functools
from unittest.mock import patch

from mockgcp.storage.client import MockClient


def mock_storage(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        with patch("google.cloud.storage.Client", MockClient):
            return func(*args, **kwargs)

    return wrapper



from __future__ import annotations

import uuid
from typing import NoReturn
from unittest import mock

from google.api_core import page_iterator
from google.cloud.exceptions import Conflict, NotFound
from google.cloud.storage.constants import _DEFAULT_TIMEOUT
from google.cloud.storage.retry import DEFAULT_RETRY

from mockgcp.storage import backend
from mockgcp.storage.bucket import MockBucket


class MockClient:
    def __init__(
        self,
        project=None,
        credentials=None,
        _http=None,
        client_info=None,
        client_options=None,
    ) -> None:
        self.project = project or "test-project-" + str(uuid.uuid1())
        self.backend = backend.StorageBackend(project=self.project)
        self.credentials = credentials
        self._http = _http
        self.client_info = client_info
        self.client_options = client_options

    @property
    def buckets(self):
        return self.backend.buckets

    @classmethod
    def create_anonymous_client(cls) -> NoReturn:
        raise NotImplementedError

    @property
    def _connection(self) -> NoReturn:
        raise NotImplementedError

    @_connection.setter
    def _connection(self, value) -> NoReturn:
        raise NotImplementedError

    def _push_batch(self, batch) -> NoReturn:
        raise NotImplementedError

    def _pop_batch(self) -> NoReturn:
        raise NotImplementedError

    def _bucket_arg_to_bucket(self, bucket_or_name):
        if isinstance(bucket_or_name, MockBucket):
            bucket = bucket_or_name
        else:
            bucket = MockBucket(self, name=bucket_or_name)
        return bucket

    @property
    def current_batch(self) -> NoReturn:
        raise NotImplementedError

    def get_service_account_email(self, project=None) -> NoReturn:
        raise NotImplementedError

    def bucket(self, bucket_name, user_project=None):
        return MockBucket(client=self, name=bucket_name, user_project=user_project)

    def batch(self) -> NoReturn:
        raise NotImplementedError

    def get_bucket(self, bucket_or_name):
        bucket = self._bucket_arg_to_bucket(bucket_or_name)

        # TODO: Use bucket.reload(client=self) when MockBucket class is implemented
        if bucket.name in self.buckets:
            return self.buckets[bucket.name]
        msg = f"404 GET https://storage.googleapis.com/storage/v1/b/{bucket.name}?projection=noAcl"
        raise NotFound(msg)

    def lookup_bucket(self, bucket_name):
        try:
            return self.get_bucket(bucket_name)
        except NotFound:
            return None

    def create_bucket(
        self,
        bucket_or_name,
        requester_pays=None,
        project=None,
        user_project=None,
        location=None,
        data_locations=None,
        predefined_acl=None,
        predefined_default_object_acl=None,
        enable_object_retention=False,
        timeout=_DEFAULT_TIMEOUT,
        retry=DEFAULT_RETRY,
    ):
        bucket = self._bucket_arg_to_bucket(bucket_or_name)
        # bucket.create(client=self, project=project)
        if bucket.name in self.buckets:
            msg = f"409 POST https://storage.googleapis.com/storage/v1/b?project={self.project}: You already own this bucket. Please select another name."
            raise Conflict(msg)
        self.buckets[bucket.name] = bucket
        self.backend.blobs[bucket.name] = {}
        return bucket

    def download_blob_to_file(
        self, blob_or_uri, file_obj, start=None, end=None
    ) -> NoReturn:
        raise NotImplementedError

    def list_blobs(
        self,
        bucket_or_name,
        max_results=None,
        page_token=None,
        prefix=None,
        delimiter=None,
        versions=None,
        projection="noAcl",
        fields=None,
    ) -> NoReturn:
        bucket = self.get_bucket(bucket_or_name)
        return bucket.list_blobs(
            max_results=max_results,
            page_token=page_token,
            prefix=prefix,
            delimiter=delimiter,
            versions=versions,
            projection=projection,
            fields=fields,
        )

    def list_buckets(
        self,
        max_results=None,
        page_token=None,
        prefix=None,
        projection="noAcl",
        fields=None,
        project=None,
    ):
        if project is None:
            project = self.project

        if isinstance(max_results, int):
            buckets = list(self.buckets.values())[:max_results]
        else:
            buckets = list(self.buckets.values())

        if isinstance(prefix, str):
            buckets = [bucket for bucket in buckets if bucket.name.startswith(prefix)]

        path = "/foo"
        page_response = {"items": buckets}
        api_request = mock.Mock(return_value=page_response)
        extra_params = {"key": "val"}

        return page_iterator.HTTPIterator(
            mock.sentinel.client,
            api_request,
            path=path,
            item_to_value=page_iterator._item_to_value_identity,
            max_results=max_results,
            page_token=mock.sentinel.token,
            extra_params=extra_params,
        )

    def create_hmac_key(
        self, service_account_email, project_id=None, user_project=None
    ) -> NoReturn:
        raise NotImplementedError

    def list_hmac_keys(
        self,
        max_results=None,
        service_account_email=None,
        show_deleted_keys=None,
        project_id=None,
        user_project=None,
    ) -> NoReturn:
        raise NotImplementedError

    def get_hmac_key_metadata(
        self, access_id, project_id=None, user_project=None
    ) -> NoReturn:
        raise NotImplementedError



from __future__ import annotations

from pathlib import Path
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING, ClassVar

if TYPE_CHECKING:
    from mockgcp.storage.blob import MockBlob
    from mockgcp.storage.bucket import MockBucket


class StorageBackend:
    _projects: ClassVar[dict[str, StorageBackend]] = {}

    def __new__(cls, project):
        if project not in cls._projects:
            cls._projects[project] = super().__new__(cls)
        return cls._projects[project]

    def __init__(self, project) -> None:
        if not hasattr(self, "project"):
            self.project = project

            self._tmp = TemporaryDirectory()
            self.data_dir = Path(self._tmp.name)
            self.buckets: dict[str, MockBucket] = {}
            self.blobs: dict[str, dict[str, MockBlob]] = {}

    def __del__(self) -> None:
        self._tmp.cleanup()



from __future__ import annotations

from datetime import datetime
from io import BytesIO
from typing import TYPE_CHECKING, BinaryIO

from google.cloud._helpers import (
    _bytes_to_unicode,
    _to_bytes,
    _rfc3339_nanos_to_datetime,
    _datetime_to_rfc3339,
)
from google.cloud.exceptions import NotFound
from google.cloud.storage._helpers import _PropertyMixin
from google.cloud.storage.constants import _DEFAULT_TIMEOUT
from google.cloud.storage.retry import (
    DEFAULT_RETRY,
    DEFAULT_RETRY_IF_GENERATION_SPECIFIED,
)

if TYPE_CHECKING:
    from pathlib import Path

    from mockgcp.storage.bucket import MockBucket


class MockBlob(_PropertyMixin):
    def __init__(
        self,
        name,
        bucket: MockBucket,
        chunk_size=None,
        encryption_key=None,
        kms_key_name=None,
        generation=None,
    ) -> None:
        name = _bytes_to_unicode(name)
        super().__init__(name)

        self.chunk_size = chunk_size  # Check that setter accepts value.
        self._bucket = bucket
        # self._acl = ObjectACL(self)
        if encryption_key is not None and kms_key_name is not None:
            msg = "Pass at most one of 'encryption_key' and 'kms_key_name'"
            raise ValueError(msg)

        self._encryption_key = encryption_key

        if kms_key_name is not None:
            self._properties["kmsKeyName"] = kms_key_name

        if generation is not None:
            self._properties["generation"] = generation

    @property
    def bucket(self) -> MockBucket:
        """Bucket which contains the object.

        :rtype: :class:`~google.cloud.storage.bucket.Bucket`
        :returns: The object's bucket.
        """
        return self._bucket

    @property
    def client(self):
        """The client bound to this blob."""
        return self.bucket.client

    @property
    def file_name(self) -> Path:
        return self.bucket.data_dir / self.name

    def exists(
        self,
        client=None,
        if_etag_match=None,
        if_etag_not_match=None,
        if_generation_match=None,
        if_generation_not_match=None,
        if_metageneration_match=None,
        if_metageneration_not_match=None,
        timeout=_DEFAULT_TIMEOUT,
        retry=DEFAULT_RETRY,
        soft_deleted=None,
    ) -> bool:
        return self.name in self.bucket.blobs

    def exists_or_raise(self, client=None, **kwargs) -> bool:
        if self.exists(client=client, **kwargs):
            return True
        msg = f"404 GET https://storage.googleapis.com/storage/v1/b/{self.bucket.name}/o/{self.name}?projection=noAcl"
        raise NotFound(msg)

    def delete(
        self,
        client=None,
        if_generation_match=None,
        if_generation_not_match=None,
        if_metageneration_match=None,
        if_metageneration_not_match=None,
        timeout=_DEFAULT_TIMEOUT,
        retry=DEFAULT_RETRY_IF_GENERATION_SPECIFIED,
    ) -> None:
        self.bucket.delete_blob(
            self.name,
            client=client,
            if_generation_match=if_generation_match,
            if_generation_not_match=if_generation_not_match,
            if_metageneration_match=if_metageneration_match,
            if_metageneration_not_match=if_metageneration_not_match,
            timeout=timeout,
            retry=retry,
        )

    def download_to_file(
        self,
        file_obj,
        client=None,
        start=None,
        end=None,
        raw_download=False,
        checksum="md5",
        **kwargs,
    ) -> None:
        file_obj.write(
            self.download_as_bytes(
                client,
                start=start,
                end=end,
                raw_download=raw_download,
                checksum=checksum,
                **kwargs,
            )
        )

    def download_to_filename(
        self,
        filename,
        client=None,
        start=None,
        end=None,
        raw_download=False,
        checksum="md5",
        **kwargs,
    ) -> None:
        with open(filename, "wb") as f:
            self.download_to_file(
                file_obj=f,
                start=start,
                end=end,
                raw_download=raw_download,
                checksum=checksum,
                **kwargs,
            )

    def download_as_bytes(
        self,
        client=None,
        start=None,
        end=None,
        raw_download=False,
        checksum="md5",
        **kwargs,
    ) -> bytes:
        self.exists_or_raise(client=client, **kwargs)
        data = self.file_name.read_bytes()
        if end is not None:
            data = data[:end]
        if start is not None:
            data = data[start:]
        return data

    def download_as_string(
        self,
        client=None,
        start=None,
        end=None,
        raw_download=False,
        checksum="md5",
        **kwargs,
    ) -> str:
        return self.download_as_text(
            client,
            start=start,
            end=end,
            raw_download=raw_download,
            checksum=checksum,
            **kwargs,
        )

    def download_as_text(
        self,
        client=None,
        start=None,
        end=None,
        raw_download=False,
        checksum="md5",
        **kwargs,
    ) -> str:
        data = self.download_as_bytes(
            client,
            start=start,
            end=end,
            raw_download=raw_download,
            checksum=checksum,
            **kwargs,
        )
        return data.decode("utf-8")

    def upload_from_file(
        self,
        file_obj: BinaryIO,
        rewind=False,
        size=None,
        content_type=None,
        num_retries=None,
        client=None,
        predefined_acl=None,
        if_generation_match=None,
        if_generation_not_match=None,
        if_metageneration_match=None,
        if_metageneration_not_match=None,
        timeout=_DEFAULT_TIMEOUT,
        checksum=None,
        retry=DEFAULT_RETRY_IF_GENERATION_SPECIFIED,
    ) -> None:
        self.file_name.parent.mkdir(parents=True, exist_ok=True)
        self.file_name.write_bytes(file_obj.read())
        self._properties["updated"] = _datetime_to_rfc3339(datetime.utcnow())
        self.bucket.blobs[self.name] = self

    def upload_from_filename(self, filename, **kwargs) -> None:
        with open(filename, "rb") as f:
            self.upload_from_file(f, **kwargs)

    def upload_from_string(self, data, **kwargs) -> None:
        data = _to_bytes(data, encoding="utf-8")
        string_buffer = BytesIO(data)
        self.upload_from_file(string_buffer, **kwargs)

    @property
    def updated(self):
        value = self._properties.get("updated")
        if value is not None:
            return _rfc3339_nanos_to_datetime(value)
